@@ -1,11 +1,11 @@
 // <algorithm> declarations  -*- C++ -*-
 
-// Copyright (C) 2007 Free Software Foundation, Inc.
+// Copyright (C) 2007, 2008, 2009, 2010 Free Software Foundation, Inc.
 //
 // This file is part of the GNU ISO C++ Library.  This library is free
 // software; you can redistribute it and/or modify it under the
 // terms of the GNU General Public License as published by the
-// Free Software Foundation; either version 2, or (at your option)
+// Free Software Foundation; either version 3, or (at your option)
 // any later version.
 
 // This library is distributed in the hope that it will be useful,
@@ -13,91 +13,20 @@
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU General Public License for more details.
 
-// You should have received a copy of the GNU General Public License along
-// with this library; see the file COPYING.  If not, write to the Free
-// Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
-// USA.
+// Under Section 7 of GPL version 3, you are granted additional
+// permissions described in the GCC Runtime Library Exception, version
+// 3.1, as published by the Free Software Foundation.
+
+// You should have received a copy of the GNU General Public License and
+// a copy of the GCC Runtime Library Exception along with this program;
+// see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
+// <http://www.gnu.org/licenses/>.
 
 /** @file bits/algorithmfwd.h
  *  This is an internal header file, included by other library headers.
  *  You should not attempt to use it directly.
  */
 
-/*
-  adjacent_find
-  binary_search
-  copy
-  copy_backward
-  count
-  count_if
-  equal
-  equal_range
-  fill
-  fill_n
-  find
-  find_end
-  find_first_of
-  find_if
-  for_each
-  generate
-  generate_n
-  includes
-  inplace_merge
-  is_heap (C++0x)
-  is_heap_until (C++0x)
-  is_sorted (C++0x)
-  is_sorted_until (C++0x)
-  iter_swap
-  lexicographical_compare
-  lower_bound
-  make_heap
-  max
-  max_element
-  merge
-  min
-  min_element
-  minmax (C++0x)
-  minmax_element (C++0x)
-  mismatch
-  next_permutation
-  nth_element
-  partial_sort
-  partial_sort_copy
-  partition
-  pop_heap
-  prev_permutation
-  push_heap
-  random_shuffle
-  remove
-  remove_copy
-  remove_copy_if
-  remove_if
-  replace
-  replace_copy
-  replace_copy_if
-  replace_if
-  reverse
-  reverse_copy
-  rotate
-  rotate_copy
-  search
-  search_n
-  set_difference
-  set_intersection
-  set_symmetric_difference
-  set_union
-  sort
-  sort_heap
-  stable_partition
-  stable_sort
-  swap
-  swap_ranges
-  transform
-  unique
-  unique_copy
-  upper_bound
-*/
-
 #ifndef _GLIBCXX_ALGORITHMFWD_H
 #define _GLIBCXX_ALGORITHMFWD_H 1
 
@@ -106,11 +35,167 @@
 #include <bits/c++config.h>
 #include <bits/stl_pair.h>
 #include <bits/stl_iterator_base_types.h>
+#include <initializer_list>
 
 _GLIBCXX_BEGIN_NAMESPACE(std)
 
+  /*
+    adjacent_find
+    all_of (C++0x)
+    any_of (C++0x)
+    binary_search
+    copy
+    copy_backward
+    copy_if (C++0x)
+    copy_n (C++0x)
+    count
+    count_if
+    equal
+    equal_range
+    fill
+    fill_n
+    find
+    find_end
+    find_first_of
+    find_if
+    find_if_not (C++0x)
+    for_each
+    generate
+    generate_n
+    includes
+    inplace_merge
+    is_heap (C++0x)
+    is_heap_until (C++0x)
+    is_partitioned (C++0x)
+    is_sorted (C++0x)
+    is_sorted_until (C++0x)
+    iter_swap
+    lexicographical_compare
+    lower_bound
+    make_heap
+    max
+    max_element
+    merge
+    min
+    min_element
+    minmax (C++0x)
+    minmax_element (C++0x)
+    mismatch
+    next_permutation
+    none_of (C++0x)
+    nth_element
+    partial_sort
+    partial_sort_copy
+    partition
+    partition_copy (C++0x)
+    partition_point (C++0x)
+    pop_heap
+    prev_permutation
+    push_heap
+    random_shuffle
+    remove
+    remove_copy
+    remove_copy_if
+    remove_if
+    replace
+    replace_copy
+    replace_copy_if
+    replace_if
+    reverse
+    reverse_copy
+    rotate
+    rotate_copy
+    search
+    search_n
+    set_difference
+    set_intersection
+    set_symmetric_difference
+    set_union
+    shuffle (C++0x)
+    sort
+    sort_heap
+    stable_partition
+    stable_sort
+    swap
+    swap_ranges
+    transform
+    unique
+    unique_copy
+    upper_bound
+  */
+
+  /**
+   * @defgroup algorithms Algorithms
+   *
+   * Components for performing algorithmic operations. Includes
+   * non-modifying sequence, modifying (mutating) sequence, sorting,
+   * searching, merge, partition, heap, set, minima, maxima, and
+   * permutation operations.
+   */
+
+  /**
+   * @defgroup mutating_algorithms Mutating Algorithms
+   * @ingroup algorithms
+   */
+
+  /**
+   * @defgroup non_mutating_algorithms Non-Mutating Algorithms
+   * @ingroup algorithms
+   */
+
+  /**
+   * @defgroup sorting_algorithms Sorting Algorithms
+   * @ingroup algorithms
+   */
+
+  /**
+   * @defgroup set_algorithms Set Operation Algorithms
+   * @ingroup sorting_algorithms
+   *
+   * These algorithms are common set operations performed on sequences
+   * that are already sorted. The number of comparisons will be
+   * linear.
+   */
+
+  /**
+   * @defgroup binary_search_algorithms Binary Search Algorithms
+   * @ingroup sorting_algorithms
+   *
+   * These algorithms are variations of a classic binary search, and
+   * all assume that the sequence being searched is already sorted.
+   * 
+   * The number of comparisons will be logarithmic (and as few as
+   * possible).  The number of steps through the sequence will be
+   * logarithmic for random-access iterators (e.g., pointers), and
+   * linear otherwise.
+   * 
+   * The LWG has passed Defect Report 270, which notes: <em>The
+   * proposed resolution reinterprets binary search. Instead of
+   * thinking about searching for a value in a sorted range, we view
+   * that as an important special case of a more general algorithm:
+   * searching for the partition point in a partitioned range.  We
+   * also add a guarantee that the old wording did not: we ensure that
+   * the upper bound is no earlier than the lower bound, that the pair
+   * returned by equal_range is a valid range, and that the first part
+   * of that pair is the lower bound.</em>
+   *
+   * The actual effect of the first sentence is that a comparison
+   * functor passed by the user doesn't necessarily need to induce a
+   * strict weak ordering relation.  Rather, it partitions the range.
+   */
+
   // adjacent_find
 
+#ifdef __GXX_EXPERIMENTAL_CXX0X__
+  template<typename _IIter, typename _Predicate>
+    bool
+    all_of(_IIter, _IIter, _Predicate);
+
+  template<typename _IIter, typename _Predicate>
+    bool
+    any_of(_IIter, _IIter, _Predicate);
+#endif
+
   template<typename _FIter, typename _Tp>
     bool 
     binary_search(_FIter, _FIter, const _Tp&);
@@ -127,6 +212,16 @@ _GLIBCXX_BEGIN_NAMESPACE(std)
     _BIter2
     copy_backward(_BIter1, _BIter1, _BIter2);
 
+#ifdef __GXX_EXPERIMENTAL_CXX0X__
+  template<typename _IIter, typename _OIter, typename _Predicate>
+    _OIter
+    copy_if(_IIter, _IIter, _OIter, _Predicate);
+
+  template<typename _IIter, typename _Size, typename _OIter>
+    _OIter
+    copy_n(_IIter, _Size, _OIter);
+#endif
+
   // count
   // count_if
 
@@ -142,13 +237,6 @@ _GLIBCXX_BEGIN_NAMESPACE(std)
     void 
     fill(_FIter, _FIter, const _Tp&);
 
-/*
-  XXX NB: return type different from ISO C++.
-  template<typename _OIter, typename _Size, typename _Tp>
-    void 
-    fill_n(_OIter, _Size, const _Tp&);
-*/
-
   template<typename _OIter, typename _Size, typename _Tp>
     _OIter
     fill_n(_OIter, _Size, const _Tp&);
@@ -165,6 +253,13 @@ _GLIBCXX_BEGIN_NAMESPACE(std)
 
   // find_first_of
   // find_if
+
+#ifdef __GXX_EXPERIMENTAL_CXX0X__
+  template<typename _IIter, typename _Predicate>
+    _IIter
+    find_if_not(_IIter, _IIter, _Predicate);
+#endif
+
   // for_each
   // generate
   // generate_n
@@ -202,6 +297,10 @@ _GLIBCXX_BEGIN_NAMESPACE(std)
     _RAIter 
     is_heap_until(_RAIter, _RAIter, _Compare);
 
+  template<typename _IIter, typename _Predicate>
+    bool
+    is_partitioned(_IIter, _IIter, _Predicate);
+
   template<typename _FIter>
     bool 
     is_sorted(_FIter, _FIter);
@@ -276,6 +375,30 @@ _GLIBCXX_BEGIN_NAMESPACE(std)
   template<typename _FIter, typename _Compare>
     pair<_FIter, _FIter>
     minmax_element(_FIter, _FIter, _Compare);
+
+  template<typename _Tp>
+    _Tp
+    min(initializer_list<_Tp>);
+
+  template<typename _Tp, typename _Compare>
+    _Tp
+    min(initializer_list<_Tp>, _Compare);
+
+  template<typename _Tp>
+    _Tp
+    max(initializer_list<_Tp>);
+
+  template<typename _Tp, typename _Compare>
+    _Tp
+    max(initializer_list<_Tp>, _Compare);
+
+  template<typename _Tp>
+    pair<_Tp, _Tp>
+    minmax(initializer_list<_Tp>);
+
+  template<typename _Tp, typename _Compare>
+    pair<_Tp, _Tp>
+    minmax(initializer_list<_Tp>, _Compare);
 #endif
 
   // mismatch
@@ -288,6 +411,12 @@ _GLIBCXX_BEGIN_NAMESPACE(std)
     bool 
     next_permutation(_BIter, _BIter, _Compare);
 
+#ifdef __GXX_EXPERIMENTAL_CXX0X__
+  template<typename _IIter, typename _Predicate>
+    bool
+    none_of(_IIter, _IIter, _Predicate);
+#endif
+
   // nth_element
   // partial_sort
 
@@ -299,6 +428,19 @@ _GLIBCXX_BEGIN_NAMESPACE(std)
     _RAIter
     partial_sort_copy(_IIter, _IIter, _RAIter, _RAIter, _Compare);
 
+  // partition
+
+#ifdef __GXX_EXPERIMENTAL_CXX0X__
+  template<typename _IIter, typename _OIter1,
+	   typename _OIter2, typename _Predicate>
+    pair<_OIter1, _OIter2>
+    partition_copy(_IIter, _IIter, _OIter1, _OIter2, _Predicate);
+
+  template<typename _FIter, typename _Predicate>
+    _FIter
+    partition_point(_FIter, _FIter, _Predicate);
+#endif
+
   template<typename _RAIter>
     void 
     pop_heap(_RAIter, _RAIter);
@@ -376,6 +518,12 @@ _GLIBCXX_BEGIN_NAMESPACE(std)
   // set_symmetric_difference
   // set_union
 
+#if defined(__GXX_EXPERIMENTAL_CXX0X__) && defined(_GLIBCXX_USE_C99_STDINT_TR1)
+  template<typename _RAIter, typename _UGenerator>
+    void
+    shuffle(_RAIter, _RAIter, _UGenerator&);
+#endif
+
   template<typename _RAIter>
     void 
     sort_heap(_RAIter, _RAIter);
@@ -392,6 +540,10 @@ _GLIBCXX_BEGIN_NAMESPACE(std)
     void 
     swap(_Tp&, _Tp&);
 
+  template<typename _Tp, size_t _Nm>
+    void
+    swap(_Tp (&)[_Nm], _Tp (&)[_Nm]);
+
   template<typename _FIter1, typename _FIter2>
     _FIter2 
     swap_ranges(_FIter1, _FIter1, _FIter2);
@@ -457,7 +609,7 @@ _GLIBCXX_BEGIN_NESTED_NAMESPACE(std, _GLIBCXX_STD_P)
     find_first_of(_FIter1, _FIter1, _FIter2, _FIter2, _BinaryPredicate);
 
   template<typename _IIter, typename _Predicate>
-    _IIter 
+    _IIter
     find_if(_IIter, _IIter, _Predicate);
 
   template<typename _IIter, typename _Funct>
@@ -468,13 +620,6 @@ _GLIBCXX_BEGIN_NESTED_NAMESPACE(std, _GLIBCXX_STD_P)
     void 
     generate(_FIter, _FIter, _Generator);
 
-/*
-  XXX NB: return type different from ISO C++.
-  template<typename _OIter, typename _Size, typename _Tp>
-    void
-    generate_n(_OIter, _Size, _Generator);
-*/
-
   template<typename _OIter, typename _Size, typename _Generator>
     _OIter
     generate_n(_OIter, _Size, _Generator);
@@ -546,7 +691,12 @@ _GLIBCXX_BEGIN_NESTED_NAMESPACE(std, _GLIBCXX_STD_P)
 
   template<typename _RAIter, typename _Generator>
     void 
-    random_shuffle(_RAIter, _RAIter, _Generator&);
+    random_shuffle(_RAIter, _RAIter,
+#ifdef __GXX_EXPERIMENTAL_CXX0X__
+		   _Generator&&);
+#else
+		   _Generator&);
+#endif
 
   template<typename _FIter, typename _Tp>
     void 
